/*

General purpose BFO board. Suitable hardware for this application is JUMA-RX1 or JUMA-TX1 DDS board. 
No display is required. However if display is connected it shows actual BFO frequency.

OH2NLT Juha Niinikoski 13.12.2005

Processor is Microchip 16F88 providing 4k words Flash, 256 bytes RAM and 256 bytes EEPROM.
Project compiled under MPLAB7.00 with Hitec PICC 8.05PL2 c-compiler


This code can be used freely for private Ham Radio projects.
Code is not a sample of clean good c language programming.
Limited amount of 16F88 Flash memory have forced to use
not clean but code size effective programming methods.


First test release 27.12.2005


*/

#define F88						// some modules use this to generate 16F88 code
#define TRX						// trx version of code, comment out for RX only

//**********************
// Include header files
//**********************

#include 	<pic.h>
#include	"jumabfo1.h"


//****************
// Hello messages
//*****************

// Note ! Display organization is 2x8 characters in one line

const char ver[] = {"OH2NLT 27.12.2005"};		// Version info
//const char lcd_ver1[] = {" OH7SV  "}; 
//const char lcd_ver2[] = {" OH2NLT "};


//******************
// Global variables
//******************

signed char tune_digit;				// which digit to tune
signed char old_tune_digit;
#define DEF_STEP 	2				// default / start up tune step
bit low_band;						// reverse tuning direction, 0 = low band

bit pwchange;						// encoder moved
signed char pwctr;					// encoder delta
unsigned char portb_copy;			// PORTB read data from irq routine. Update intervall 1ms
			
bit dmode;							// LCD display mode, only two alternatives
bit testmode;						// test mode indicator

int ms_timer;						// millisecond timer, decremented by irq routine, stop at 0

union								// Battery voltage from A/D converter
	{
	int battery;					// A/D counts = U/2
	unsigned char adcregs[2];
	} ad;

union								// Tuning pot(J3 0-5V) from from A/D converter
	{
	int pot;
	unsigned char potregs[2];
	} pot;

#define POT_OFFSET	512				// Tuning potentiometer offset

unsigned char smeter;				// S-meter data from A/D converter. Smax = 240mV

// VFO structure
	union							// VFO data
	{
	unsigned long vfos[6];			// VFO A = 0, B = 1
	unsigned char vfo_bytes[24];
	}bank1 vfo;

char vfo_idx;						// index to VFO table
#define vfo_A	0					// address in VFO table
#define vfo_B	2
#define lo1		4					// lo frequency constant

#define hi_VFO	1					// table offset for high band

//#define vfo_A			(vfo.vfos[0])	// define VFOs at structures
//#define vfo_B			(vfo.vfos[1])

#define	disp_freq		(m1.m1l)	// define display freq storage at large mul structures
long dds_freq;						// dds tuning word frequency
long tune_step;						// tuning step

// tuning limits
#define TUNE_LOW_LIMIT	0			// test setup can go down to 0 Hz
#define TUNE_HIGH_LIMIT 0x6666666	// 8 MHz upper limit
#define TUNE_TESTH_LIMIT 132875550	// 9,9 MHz upper limit for LO1 freq

// tuning steps
#define TUNE_STEP_LOW 	134			// 10 Hz
#define TUNE_STEP_MID	1342		// 100 Hz
//#define TUNE_STEP_MID	6711		// 500 Hz
#define TUNE_STEP_UP	13422		// 1 kHz
#define TUNE_STEP_HI	1342177		// 100 kHz

// tuning other
#define TX_SHIFT		9395		// 700 Hz, when TX subtract this

#define LO1_LO			0x733332C	// 9.000.000 Hz first LO for low band
#define LO1_HI			0x733332C	// 9.000.000 Hz first LO for high band

// user interface timers
#define SAVE_DELAY		2000		// button press time (ms) for save frequency

// some DDS constants
#define BITS28		268435456UL		// 2^28 constant
#define BITS32		4294967296UL	// 2^32 constant
#define DDC_CLK		20000000UL		// DDS Chip clock

#define DMULT		320000000UL		// display multiplier, 0,07450581Hz/bit * 2^32

#define START_FREQ	0x3FFFFFC		// factory default (empty EEPROM) frequency, 5.000.000 Hz

#define MHz_cal		0xCCCCCC		// exact 1.000.000 Hz setup for calibration

#define FILTER_SEL	(0x3333333 >> 14)	// Filter select 0x3333333 = 4.000.000 Hz

// fixed BFO frequencies

#define KHz_134300	1802544			// J1-1 closed, J1-2 closed
#define KHz_131300	1762279			// J1-1 closed, J1-2 open
#define KHz_132800	1782411			// J1-1 open, J1-2 don't care

// J1 signals
#define J1_1		(1<<4)			// F3
#define J1_2		(1<<5)			// F1/F2
#define J1_3		(1<<6)			//
#define J1_4		(1<<7)			//

#define TUNE_SEL	RA5				// enable pot tune if 0V

// other
bit	aux_out;						// aux output bit. Transferred to output with DDS write.				

//*****************
// Include C files
//*****************

#include	"ad9833.c"
#include	"large_mul.c"
#include	"dcrxtim.c"
#include	"dcrxlcd.c"
#include 	"dcrxadc.c"


//******************
// Processor config
//******************

	__CONFIG(UNPROTECT & HS & WDTDIS & PWRTEN & BOREN & LVPDIS & MCLRDIS);	// 16F88 config
//	__CONFIG(2, FCMDIS & IESODIS);
// CONFIG2 does not work with 805PL2 compiler, patch manually

	__IDLOC(0007);				// Device ID


//******************
// Helper functions
//******************


// generate tone RA0 pin (battery voltage measurement)

void tone(int pitch, int len)		// pitch = 1/2 cycle in us units, len = ms
	{
	ms_timer = len;
	while(ms_timer != 0)
		{
		TONE_OUT = 1;
		delayu(pitch);
		TONE_OUT = 0;
		delayu(pitch);
		}
	}

// tone definitions

#if 0
#define LOW_TONE	2500	// 200 Hz
#define MID_TONE	833		// 600 Hz
#define HIGH_TONE	500		// 1 kHz
#endif

#define LOW_TONE	1911	// C 261,6 Hz
#define MID_TONE	1517	// E 329,6 Hz
#define HIGH_TONE	1276	// G 392 Hz
#define HIGH2_TONE	1012	// B 493,8 Hz

#define CHANGE_LEN	100		// step change beep length (ms)


#define MORSE_TONE	955		// 523,2 Hz
#define DI 0
#define DAA 1
#define UNIT 40				// morse unit length (ms)


// generate morse symbol

void beep(char l)
	{
	if(l == DI)
		tone(MORSE_TONE, UNIT);
	else
		tone(MORSE_TONE, (3*UNIT));
	delaym(UNIT);
	}

// say CW hello 

void morse_message(void)		// hello experiment
	{
	TONE_DIR = 0;				// make tone output active

	beep(DI);					// say JUMA
	beep(DAA);
	beep(DAA);
	beep(DAA);
	delaym(3*UNIT);
	beep(DI);
	beep(DI);
	beep(DAA);
	delaym(3*UNIT);
	beep(DAA);	
	beep(DAA);
	delaym(3*UNIT);
	beep(DI);
	beep(DAA);
	delaym(7*UNIT);				// word space

	beep(DAA);					// say BFO
	beep(DI);
	beep(DI);
	beep(DI);
	delaym(3*UNIT);
	beep(DI);
	beep(DI);
	beep(DAA);
	beep(DI);
	delaym(3*UNIT);
	beep(DAA);
	beep(DAA);
	beep(DAA);

	TONE_DIR = 1;				// return to input mode
//	delaym(7*UNIT);				// end delay
	}

void morse_roger(void)			// say roger
	{
	TONE_DIR = 0;				// make tone output active

	beep(DI);					// say R
	beep(DAA);
	beep(DI);

	TONE_DIR = 1;				// return to input mode
	}

// store & read frequency to/from internal EEPROM
// EEPROM[0...23] = frequency, EEPROM[32] = step
void save_freq(void)
	{
	char x;
	for(x=0; x<24; x++)
		{
		EEPROM_WRITE(x, (vfo.vfo_bytes[x]));		// save VFO data
		}
	EEPROM_WRITE(32, tune_digit);					// save tune digit
	}

void read_freq(void)
	{
	char x;
	for(x=0; x<24; x++)
		{
		vfo.vfo_bytes[x] = EEPROM_READ(x);			// restore VFO data
		}
	tune_digit = EEPROM_READ(32);					// restore tuning step
	}


// check rx_frequency (DDS tuning word) limits and
// force rx_freq into limits
// status = 0 if ok, 1 if low, 2 if high

char check_freq_limits(void)
	{
	if((long)dds_freq < TUNE_LOW_LIMIT)		// limitters
		{
		dds_freq = TUNE_LOW_LIMIT;			// low = 0
		return(1);
		}
	if(testmode)
		{
		if(dds_freq > TUNE_TESTH_LIMIT)
			{
			dds_freq = TUNE_TESTH_LIMIT;			// high = 9,9 MHz
			return(2);
			}
		}
	else
		{
		if(dds_freq > TUNE_HIGH_LIMIT)
			{
			dds_freq = TUNE_HIGH_LIMIT;			// high = 8 MHz
			return(2);
			}
		}

	return(0);
	}

// Generate beep tone according tune step 

const int tune_tone[5] = {0, LOW_TONE, MID_TONE, HIGH_TONE, HIGH2_TONE};	// 16F88 tune step tone indicators

void step_beep( void )
	{
	TONE_DIR = 0;								// make tone output active

	tone(tune_tone[tune_digit], CHANGE_LEN);	// select tone & beep

	TONE_DIR = 1;								// return to input mode
	} 


//*****************
// Main function
//*****************

void main(void)
{
/* Init IO Ports & software modules */

//Set up ADC and PORTs

	ANSEL = ANSEL_INIT;		// A/D converter pin select

	PORTA = PORTA_INIT;		// A/D & control signals
	PORTB = PORTB_INIT;		// LCD & SW port

	ADCON1 = ADCON1_INIT;	// PORTA = digital I/O during start

	TRISA = TRISA_INIT;		// set port direction bits
	TRISB = TRISB_INIT;

	aux_out = 0;

	Init_timers();			// Init timers & interrupt system
	lcdinit();				// Init LCD
	init_ad9833();			// init DDS
//	read_freq();			// read VFO start up values from EEPROM

// band & VFO select logic

	dds_freq = KHz_132800;	// set frequency in use
	tune_ad9833(dds_freq);	// set DDS

	tune_digit = DEF_STEP;
	old_tune_digit = tune_digit;

	lcdinit();				// Init LCD

	putst("JUMA-BFO");		// Show RX type & version
	lcdcom(SECOND_HALF);
	putst("1  v0.01");

	delaym(200);			// wait for audio ready
	pwctr = 0;				// clear encoder
	morse_message();		// say audible hello

	lcdclr( );
	putst(" OH7SV  ");		// Tell who we are
	lcdcom(SECOND_HALF);
	putst(" OH2NLT ");
	delaym(1800);			// "optimized" message display time

	set_chgen();			// download soft characters
	lcdclr( );				// clear display


#if 0
	if((portb_copy & BUTTON_1) == 0) // Oscillator test ?
		{
		putst("1 MHz");		// show calibration message
		tune_ad9833(MHz_cal);	// set 1.000.000 MHz
		dds_freq = MHz_cal;
		delaym(10000);		// show message 10s
		testmode = 1;		// set test mode flag
		}
#endif

// Free running Main Control Loop

	for(;;)
	{
// do A/D measurements

	measure_all();								// measure & store A/D channells


// select frequency
	if((portb_copy & J1_1) == 0)				// evaluate J1-2 state
		{
		if((portb_copy & J1_2) == 0)
			dds_freq = KHz_134300;				// J1-2 closed set F1
		else
			dds_freq = KHz_131300;				// J1-2 open set F2
		}
	else
		dds_freq = KHz_132800;					// all open, set F3

// potentiometer fine tuning

	if(TUNE_SEL == 0)							// about +/- 1kHz offset tuning
		dds_freq = dds_freq + ((pot.pot - POT_OFFSET) * 26);

// set DDS

		tune_ad9833(dds_freq);					// set DDS


// display selected frequency if display is connected

	lcdcom(FIRST_HALF);							// select line one (first 8 chr) display

//	draw_s_meter(smeter);						// draw smeter graph, scale = 0...48 counts

	if(TUNE_SEL == 0)							// potentiometer offset tuning in use
		{
		putvolt(pot.pot, 0);						// raw A/D count
		putst(" #  ");		
		}
	else										// show voltage in fixed freq mode
		{	
		putvolt(ad.battery << 1, 1);				// A/D count * 2 = voltage  12.00V
		putst(" V ");
		}


// calculate and display DDS frequency in use (dds_freq)

	disp_freq = dds_freq;						// in test mode display selected LO1 constant	
	lcdcom(SECOND_HALF);						// line two allways frequency display
//	m1.m1l = disp_freq;							// calculate display frequency from DDS setup word
	m2.m2l = DMULT;
	vlmul();									// very large multiply, 32*32 bits
	add_prod(3, 0x80);							// round, add 2^32/2 to product before truncation 26.01.2005
	putfreq(prod.prodlng.prodhi);				// display frequency
	}											// end for
}												// end main
